@@ -2,7 +2,6 @@ import React from 'react';
 import { Link } from 'react-router-dom';
 import PropTypes from 'prop-types';
 import { useTranslation } from 'react-i18next';
-import './LogsSearchLink.css';
 import { getLogsUrlParams } from '../../helpers/helpers';
 import { MENU_URLS } from '../../helpers/constants';
 
@@ -13,11 +12,16 @@ const LogsSearchLink = ({
 
     const to = link === MENU_URLS.logs ? `${MENU_URLS.logs}${getLogsUrlParams(search && `"${search}"`, response_status)}` : link;
 
-    return <Link to={to}
-                 className={'stats__link'}
-                 tabIndex={0}
-                 title={t('click_to_view_queries')}
-                 aria-label={t('click_to_view_queries')}>{children}</Link>;
+    return (
+        <Link
+            to={to}
+            tabIndex={0}
+            title={t('click_to_view_queries')}
+            aria-label={t('click_to_view_queries')}
+        >
+            {children}
+        </Link>
+    );
 };
 
 LogsSearchLink.propTypes = {
